/* $Id: vll.c,v 1.1 2000/01/12 23:56:32 bjj Exp $
 *
 * Implementation of the transmit side VLL protocol for legOS.
 * This can be used to directly control or program the MicroScout
 * or the Code Pilot.  The RCX is not capable of transmitting as
 * fast as the Scout, probably because of the type of LED in the
 * light sensor.  At full speed, the RCX cannot drive the MicroScout
 * motor constantly in direct mode.
 *
 * This code borrows heavily from MicroScoutVLL.ps (GPL) by
 * Doug Eaton <doug@eaton.dhs.org>.  He has a web page about VLL
 * "simulation" at http://eaton.dhs.org/lego/ and did all of the
 * work reverse-engineering the MicroScout/Code Pilot barcodes.
 *
 * Usage:
 *	vll_init(&SENSOR_1);	// light sensor on port one is xmitter
 *	vll_send(VLL_MS_FWD);
 *	...
 *	vll_stop();		// if you want the light off
 *
 * Make sure vll_send() is called from a very high priority task if you
 * are multitasking.  If another task alters its timing the send will fail.
 */

#include <unistd.h>
#include <time.h>
#include <dsensor.h>

/* From Doug Eaton's reverse-engineering */
#define VLL_CHECKSUM(n)	((7 - ((n) + (n)/4 + (n)/16)) & 0x7)

/* Sensor hardware index, select with vll_init() */
static unsigned char vll_out;

/****************************************************************************
 * NOTE  NOTE  NOTE  NOTE  NOTE  NOTE  NOTE  NOTE  NOTE  NOTE NOTE NOTE NOTE
 *
 * I wrote this for legOS 0.2.3, which requires a patch to dsensor.c to
 * avoid interfering with my direct control of the sensor power output.
 * If you don't apply this patch the light sensor's LED will stay dark
 * even when you attempt to transmit VLL.
 ****************************************************************************/
extern unsigned char ds_ignore;

/* These do get inlined.  I'm not sure if clobbering r0 is legal and the
 * compiler would protect it if needed, or if clobbering r0 is flirting with
 * disaster.  It doesn't matter much since the generated code always smashes
 * r0 immediately after this to make the first argument to msleep().
 */
static inline void
VLL_SPACE()
{
	__asm__("
	mov.b @_vll_out, r0l
	bset r0l, @0xbb:8
	");
}

static inline void
VLL_BAR()
{
	__asm__("
	mov.b @_vll_out, r0l
	bclr r0l, @0xbb:8
	");
}

/* Make each VLL code execute in about 0.400 seconds.  This value seems
 * to work nearly 100%, while slightly smaller values like 375ms drop off
 * to near 0% success.
 */
const int vll_pulsewidth = 400 / 35;

/* If you transmit two VLL messages back to back, a silent period of
 * vll_gapwidth will be inserted.  Anything much smaller than this will
 * result in transmit failures (MicroScout makes a low bzzt sound).
 */
#define vll_gapwidth (vll_pulsewidth * 8)

#define VLL_LONG()	msleep(vll_pulsewidth * 2)
#define VLL_SHORT()	msleep(vll_pulsewidth)

#define VLL_LONG_BAR() do { VLL_BAR(); msleep(vll_pulsewidth*2); } while (0)
#define VLL_SHORT_BAR() do { VLL_BAR(); msleep(vll_pulsewidth); } while (0)

/* If you flicker sensor 1 on and off with 6/5 of the cycles bright, a
 * second sensor directly across from it reads about even brightness.
 * This is probably due to the time it takes to warm up the LED fillament.
 * The LEGO light sensor red LED is clearly not designed for signalling
 * applications!
 */
#define VLL_LONG_SPACE() do { VLL_SPACE(); msleep(12*vll_pulsewidth/5); } while (0)
#define VLL_SHORT_SPACE() do { VLL_SPACE(); msleep(6*vll_pulsewidth/5); } while (0)

static time_t last_vll_send;

void
vll_init(volatile unsigned *sensor)
{
	if (sensor == &SENSOR_3)
		vll_out = 0;
	else if (sensor == &SENSOR_2)
		vll_out = 1;
	else if (sensor == &SENSOR_1)
		vll_out = 2;

	ds_passive(sensor);
	ds_rotation_off(sensor);
	ds_ignore |= 1 << vll_out;
	last_vll_send = 0;
	VLL_SPACE();
}

void
vll_stop()
{
	time_t now = sys_time;

	if (now - last_vll_send < vll_gapwidth)
		msleep(vll_gapwidth - (now - last_vll_send));

	ds_ignore &= ~(1 << vll_out);
}

void
vll_send(char code)
{
	int value = VLL_CHECKSUM(code);
	time_t now = sys_time;
	int i;

	if (now - last_vll_send < vll_gapwidth)
		msleep(vll_gapwidth - (now - last_vll_send));

	// start bit
	VLL_SHORT_BAR();

	value = (value << 7) | code;
	for (i = 9; i >= 0; --i) {
		if (value & (1 << i)) {
			VLL_SHORT_SPACE();
			VLL_LONG_BAR();
		} else {
			VLL_LONG_SPACE();
			VLL_SHORT_BAR();
		}
	}
	// stop bit
	VLL_SHORT_SPACE();
	VLL_SHORT_BAR(); VLL_LONG();

	VLL_SPACE();
	last_vll_send = sys_time;
}
